#include "SceneModel.h"
#include "ModelLoader.h"
#include <glm\gtc\matrix_transform.hpp>
#include "../PhysxManager.h"

using namespace babyOnWheels;
using namespace scene;
using namespace physx;

SceneModel::SceneModel(glm::mat4& matrix)
	:SceneObject(matrix)
{
}

SceneModel::SceneModel(SceneModel* _model, PhysxManager* physxManager)
	:SceneObject(_model->modelMatrix)
{
	indices = _model->indices;
	texture_handle = _model->texture_handle;
	vertexbuffer = _model->vertexbuffer;
	uvbuffer = _model->uvbuffer;
	normalbuffer = _model->normalbuffer;
	elementbuffer = _model->elementbuffer;

	model_mat = _model->modelMatrix;

	if (_model->collider != nullptr) {
		collider = physxManager->createStaticCollider(modelMatrix, _model->collider->size);
	}
	if (_model->body != nullptr) {
		body = physxManager->createRigidbody(modelMatrix, _model->body->collider->getMass(), _model->body->size);
	}

	for (const auto model : _model->sceneModels_Map) {
		sceneModels_Map.insert(std::pair<string, SceneModel*>(model.first, new SceneModel(model.second, physxManager)));
	}
}


SceneModel::~SceneModel()
{
	glDeleteBuffers(1, &vertexbuffer);
	glDeleteBuffers(1, &uvbuffer);
	glDeleteBuffers(1, &normalbuffer);
	glDeleteBuffers(1, &elementbuffer);
}

void SceneModel::draw(GLuint handle)
{
	glm::mat4 ModelMatrix = glm::rotate(modelMatrix, glm::radians(90.0f), glm::vec3(1, 0, 0));
	glUniformMatrix4fv(handle, 1, GL_FALSE, &ModelMatrix[0][0]);
	//textur binden
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texture_handle);
	// 1rst attribute buffer : vertices
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
	glVertexAttribPointer(
		0,                  // attribute
		3,                  // size
		GL_FLOAT,           // type
		GL_FALSE,           // normalized?
		0,                  // stride
		(void*)0            // array buffer offset
		);

	// 2nd attribute buffer : UVs
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, uvbuffer);
	glVertexAttribPointer(
		1,                                // attribute
		2,                                // size
		GL_FLOAT,                         // type
		GL_FALSE,                         // normalized?
		0,                                // stride
		(void*)0                          // array buffer offset
		);

	// 3rd attribute buffer : normals
	glEnableVertexAttribArray(2);
	glBindBuffer(GL_ARRAY_BUFFER, normalbuffer);
	glVertexAttribPointer(
		2,                                // attribute
		3,                                // size
		GL_FLOAT,                         // type
		GL_FALSE,                         // normalized?
		0,                                // stride
		(void*)0                          // array buffer offset
		);

	// Index buffer
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementbuffer);

	// Draw the triangles !
	glDrawElements(
		GL_TRIANGLES,      // mode
		indices.size(),    // count
		GL_UNSIGNED_SHORT,   // type
		(void*)0           // element array buffer offset
		);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);
	glDisableVertexAttribArray(2);

	//DRAW ALL CHILDREN
	for (const auto model : sceneModels_Map) {
		model.second->draw(handle);
	}
}

void SceneModel::update(float time_delta)
{
	if (body != nullptr) {
		modelMatrix = PhysxManager::getShapePosition(body->collider);
	}

	for (const auto model : sceneModels_Map) {
		model.second->update(time_delta);
	}
}

void SceneModel::translate(glm::vec3 _value) {
	modelMatrix = glm::translate(modelMatrix, _value);
	if (collider) {
		collider->collider->setGlobalPose(PxTransform(PhysxManager::convertMatToPhysx(modelMatrix)));
	}
	if (body) {
		body->collider->setGlobalPose(PxTransform(PhysxManager::convertMatToPhysx(modelMatrix)));
	}
	for (const auto model : sceneModels_Map) {
		model.second->translate(_value);
	}
}
void SceneModel::rotate(float angle, glm::vec3 _axis){
	modelMatrix = glm::rotate(modelMatrix, angle, _axis);
	if (collider) {
		collider->collider->setGlobalPose(PxTransform(PhysxManager::convertMatToPhysx(modelMatrix)));
	}
	if (body) {
		body->collider->setGlobalPose(PxTransform(PhysxManager::convertMatToPhysx(modelMatrix)));
	}
	for (const auto model : sceneModels_Map) {
		model.second->modelMatrix = glm::translate(glm::mat4(1), -glm::vec3(modelMatrix[3])) * model.second->modelMatrix;
		model.second->modelMatrix = glm::rotate(glm::mat4(1), angle, _axis) * model.second->modelMatrix;
		model.second->modelMatrix = glm::translate(glm::mat4(1), glm::vec3(modelMatrix[3])) * model.second->modelMatrix;
		//model.second->rotate(angle, _axis);
	}
}


void SceneModel::loadSceneModel(const char* path, GLuint _texture_handle) {
	texture_handle = _texture_handle;
	// Read our .obj file
	indexed_vertices;
	std::vector<glm::vec2> indexed_uvs;
	std::vector<glm::vec3> indexed_normals;	
	ModelLoader* loader = new ModelLoader();
	bool res = loader->loadAssImp(path, indices, indexed_vertices, indexed_uvs, indexed_normals);

	// Load it into a VBO

	
	glGenBuffers(1, &vertexbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
	glBufferData(GL_ARRAY_BUFFER, indexed_vertices.size() * sizeof(glm::vec3), &indexed_vertices[0], GL_STATIC_DRAW);

	
	glGenBuffers(1, &uvbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, uvbuffer);
	glBufferData(GL_ARRAY_BUFFER, indexed_uvs.size() * sizeof(glm::vec2), &indexed_uvs[0], GL_STATIC_DRAW);

	
	glGenBuffers(1, &normalbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, normalbuffer);
	glBufferData(GL_ARRAY_BUFFER, indexed_normals.size() * sizeof(glm::vec3), &indexed_normals[0], GL_STATIC_DRAW);

	// Generate a buffer for the indices as well
	
	glGenBuffers(1, &elementbuffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementbuffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(unsigned short), &indices[0], GL_STATIC_DRAW);
}

bool SceneModel::checkIfInside(glm::mat4& _other) {
	
	glm::vec3 scale = glm::vec3(trigger[0][0], trigger[1][1], trigger[2][2]);
	glm::vec3 pos = glm::vec3(trigger[3]);

	//glm::vec3 otherScale = glm::vec3(_other[0][0], _other[1][1], _other[2][2]);
	glm::vec3 otherPos = glm::vec3(_other[3]);

	if (otherPos.x > pos.x - scale.x / 2 && otherPos.x < pos.x + scale.x / 2 &&
		otherPos.y > pos.y - scale.y / 2 && otherPos.y < pos.y + scale.y / 2 &&
		otherPos.z > pos.z - scale.z / 2 && otherPos.z < pos.z + scale.z / 2) {

		return true;
	}
	return false;
}

glm::vec4 SceneModel::getSphere()
{
	glm::vec3 vert = indexed_vertices[0] * glm::vec3(model_mat[3]);

	glm::vec3 min = vert;
	glm::vec3 max = vert;

	glm::vec3 center;
	float radius;

	for (int i = 1; i < indexed_vertices.size(); i++)
	{
		glm::vec3 vert = indexed_vertices[i] * glm::vec3(model_mat[3]);

		if (min.x > vert.x)
		{
			min.x = vert.x;
		}

		if (min.y > vert.y)
		{
			min.y = vert.y;
		}

		if (min.z > vert.z)
		{
			min.z = vert.z;
		}

		if (max.x < vert.x)
		{
			max.x = vert.x;
		}

		if (max.y < vert.y)
		{
			max.y = vert.y;
		}

		if (max.z < vert.z)
		{
			max.z = vert.z;
		}
	}
	
	center.x = max.x - ((max.x - min.x) / 2);
	center.y = max.y - ((max.y - min.y) / 2);
	center.z = max.z - ((max.z - min.z) / 2);

	radius = (max.x - min.x) / 2;
	if (radius < ((max.y - min.y) / 2)){
		radius = ((max.y - min.y) / 2);
	}
	if (radius < ((max.z - min.z) / 2)){
		radius = ((max.z - min.z) / 2);
	}


	glm::vec3 scale = glm::vec3(trigger[0][0], trigger[1][1], trigger[2][2]);
	glm::vec3 pos = glm::vec3(trigger[3]);

	// return glm::vec4(center, radius);

	float max_scale = 0;
	if (scale.x > max_scale)
	{
		max_scale = scale.x;
	}
	else if (scale.y > max_scale)
	{
		max_scale = scale.y;
	}
	else if (scale.z > max_scale)
	{
		max_scale = scale.z;
	}

	return glm::vec4(pos, max_scale);


}